/*
 * This file is part of firk's Common C Library
 *  qprint.c - qlog-compatible formatted string printing
 *
 * (c) firk 2019, 2020, 2021, 2023; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include "types.h"
#include "string.h"
#include "malloc.h"
#include "qlog.h"
#define FCL_WITH_VA_LIST
#include "qprint.h"

extern int qp_callback_allowed(void) {
#ifndef FCL_NO_QPRINT_CB
  return 1;
#else
  return 0;
#endif
}

#define FCL_STRFMT_WITH_TIMESTR
#define PA qprint_ctx * ctx,
#define CA ctx,
#define CHECK_STOP if(ctx->fail) return 0;
#define remlen ctx->rl
#define maxlen ctx->ml
#define maxlen_present ctx->mlp

static int write_all(qprint_ctx * ctx, char const * buf, size_t nb) {
  size_t csz;
  ssize_t wsz;
  if(nb) {
#ifndef FCL_NO_QPRINT_CB
    if(ctx->fd==-2) return ctx->cb(ctx->cbdata, buf, nb);
#endif
    if(ctx->fd<0) { errno = EBADF; return -1; }
    do {
      csz = nb;
      while(((ssize_t)csz)<=0) csz/=2;
      wsz = write(ctx->fd, buf, csz);
      if(wsz<0) {
        if(errno==EINTR) continue;
        return -1;
      }
      assert(wsz>=1 && ((size_t)wsz)<=csz && csz<=nb);
      buf += wsz;
      nb -= wsz;
    } while(nb);
  }
  return 0;
}

extern int qp_flush(qprint_ctx * ctx) {
  if(ctx->fail) return -1;
  if(ctx->bufpos) {
    if(write_all(ctx, ctx->buf, ctx->bufpos)<0) ctx->fail = errno?errno:-1;
    ctx->bufpos = 0;
  }
  return ctx->fail?-1:0;
}

extern int qp_softflush(qprint_ctx * ctx) {
  if(ctx->fail) return -1;
  if(ctx->fd==-1) return 0;
  if(ctx->bufpos) {
    if(write_all(ctx, ctx->buf, ctx->bufpos)<0) ctx->fail = errno?errno:-1;
    ctx->bufpos = 0;
  }
  return ctx->fail?-1:0;
}

/* NOTE that fd!=-1 only allowed if bufsize>0 && buf!=NULL
 * assert(fd==-1 || bufsize && buf); */
static void E_write(PA char const * s, size_t len) {
  size_t add;
  /* not allowing (bufpos==bufsize && fd!=-1 && fail) because
   * qp_flush() resetting bufpos even on failure */
  assert(ctx->bufpos<ctx->bufsize || ctx->bufpos==ctx->bufsize && ctx->fd==-1);
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  if(ctx->fail) return;
  if(!len) return;
  if(len<=ctx->bufsize-ctx->bufpos) {
    if(ctx->buf) bcopy(s, ctx->buf+ctx->bufpos, len);
    ctx->bufpos += len;
    if(ctx->bufpos==ctx->bufsize && ctx->fd!=-1) qp_flush(ctx);
  } else if(len<ctx->bufsize/2 || ctx->fd==-1) {
    add = ctx->bufsize-ctx->bufpos;
    if(ctx->buf) bcopy(s, ctx->buf+ctx->bufpos, add);
    ctx->bufpos = ctx->bufsize;
    if(ctx->fd==-1) { ctx->fail = ENOSPC; return; }
    qp_flush(ctx);
    if(!ctx->fail) bcopy(s+add, ctx->buf, ctx->bufpos=len-add);
  } else {
    qp_flush(ctx);
    if(!ctx->fail && write_all(ctx, s, len)<0) ctx->fail = errno?errno:-1;
  }
}

static void E_repeat(PA char c, size_t len) {
  size_t add;
  assert(ctx->bufpos<ctx->bufsize || ctx->bufpos==ctx->bufsize && ctx->fd==-1);
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  if(ctx->fail) return;
  if(!len) return;
  if(len<=ctx->bufsize-ctx->bufpos) {
    if(ctx->buf) memset(ctx->buf+ctx->bufpos, c, len);
    ctx->bufpos += len;
    if(ctx->bufpos==ctx->bufsize && ctx->fd!=-1) qp_flush(ctx);
  } else {
    add = ctx->bufsize-ctx->bufpos;
    if(ctx->buf) memset(ctx->buf+ctx->bufpos, c, add);
    ctx->bufpos = ctx->bufsize;
    if(ctx->fd==-1) { ctx->fail = ENOSPC; return; }
    qp_flush(ctx);
    if(ctx->fail) return;
    ctx->bufpos = len-add;
    if(ctx->bufpos<ctx->bufsize) {
      memset(ctx->buf, c, ctx->bufpos);
    } else {
      memset(ctx->buf, c, ctx->bufsize);
      while(ctx->bufpos>=ctx->bufsize) {
        if(write_all(ctx, ctx->buf, ctx->bufsize)<0) {
          ctx->fail = errno?errno:-1;
          ctx->bufpos = 0;
          return;
        }
        ctx->bufpos -= ctx->bufsize;
      }
    }
  }
}

static void E_print_char(PA char c) {
  assert(ctx->bufpos<ctx->bufsize || ctx->bufpos==ctx->bufsize && ctx->fd==-1);
  if(remlen) remlen--;
  if(maxlen_present) { if(maxlen) maxlen--; else return; }
  if(ctx->fail) return;
  if(ctx->bufpos==ctx->bufsize && ctx->fd==-1) { ctx->fail = ENOSPC; return; }
  assert(ctx->bufpos<ctx->bufsize);
  if(ctx->buf) ctx->buf[ctx->bufpos] = c;
  ctx->bufpos++;
  if(ctx->bufpos==ctx->bufsize && ctx->fd!=-1) qp_flush(ctx);
}

#include "strfmt.inc"

static char qp_outbuf[4096], qp_errbuf[4096];
qprint_ctx qp_out = { .fd=1, .buf=qp_outbuf, .bufsize=sizeof(qp_outbuf) };
qprint_ctx qp_err = { .fd=2, .buf=qp_errbuf, .bufsize=sizeof(qp_errbuf) };

extern void qp_init_buf(qprint_ctx * ctx, char * dest, size_t destsz) {
  ctx->fd=-1; ctx->fail=0; ctx->buf=dest; ctx->bufsize=destsz;
  ctx->bufpos=0; ctx->rl=0; ctx->mlp=0;
}

extern int qp_printf(qprint_ctx * ctx, char const * fmt, ...) {
  va_list arg;
  if(ctx->fail) return -1;
  va_start(arg, fmt);
  if(E_vprint(ctx, fmt, arg)<0) {
    qlog("qp_printf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  va_end(arg);
  return ctx->fail?-1:0;
}

extern int qp_vprintf(qprint_ctx * ctx, char const * fmt, va_list arg) {
  if(ctx->fail) return -1;
  if(E_vprint(ctx, fmt, arg)<0) {
    qlog("qp_vprintf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  return ctx->fail?-1:0;
}

extern int qp_bvprintf(char const * title, qprint_ctx * ctx,
                       char const * fmt, va_list arg) {
  if(ctx->fail) return -1;
  if(E_vprint(ctx, fmt, arg)<0) {
    qlog("%s: bad format string: \"%S\"", title, fmt);
    exit(-1);
  }
  return ctx->fail?-1:0;
}

extern int qp_snprintf(char * dest, size_t destsize, char const * fmt, ...) {
  qprint_ctx ctx;
  va_list arg;
  if(!dest || !destsize) return -1;
  ctx.fd=-1;ctx.fail=0;ctx.buf=dest;ctx.bufsize=destsize-1;ctx.bufpos=0;
  va_start(arg, fmt);
  if(E_vprint(&ctx, fmt, arg)<0) {
    qlog("qp_snprintf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  va_end(arg);
  assert(ctx.bufpos<destsize);
  ctx.buf[ctx.bufpos] = 0;
  return ctx.fail?-1:0;
}

extern int qp_vsnprintf(char * d, size_t dsz, char const * fmt, va_list arg) {
  qprint_ctx ctx;
  if(!d || !dsz) return -1;
  ctx.fd=-1;ctx.fail=0;ctx.buf=d;ctx.bufsize=dsz-1;ctx.bufpos=0;
  if(E_vprint(&ctx, fmt, arg)<0) {
    qlog("qp_vsnprintf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  assert(ctx.bufpos<dsz);
  ctx.buf[ctx.bufpos] = 0;
  return ctx.fail?-1:0;
}

extern size_t qp_strnlenf(size_t maxl, char const * fmt, ...) {
  qprint_ctx ctx;
  va_list arg;
  if(!maxl) return 0;
  ctx.fd=-1;ctx.fail=0;ctx.buf=NULL;ctx.bufsize=maxl;ctx.bufpos=0;
  va_start(arg, fmt);
  if(E_vprint(&ctx, fmt, arg)<0) {
    qlog("qp_strnlenf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  va_end(arg);
  assert(ctx.bufpos<=maxl);
  return ctx.bufpos;
}

extern size_t qp_vstrnlenf(size_t maxl, char const * fmt, va_list arg) {
  qprint_ctx ctx;
  if(!maxl) return 0;
  ctx.fd=-1;ctx.fail=0;ctx.buf=NULL;ctx.bufsize=maxl;ctx.bufpos=0;
  if(E_vprint(&ctx, fmt, arg)<0) {
    qlog("qp_vstrnlenf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  assert(ctx.bufpos<=maxl);
  return ctx.bufpos;
}

extern int qp_putc(PA char c) {
  E_print_char(CA c);
  return ctx->fail?-1:0;
}

extern int qp_putce(PA char c) {
  E_print_escaped_char(CA c);
  return ctx->fail?-1:0;
}

extern int qp_puts(PA char const * s) {
  E_print_str(CA s);
  return ctx->fail?-1:0;
}

extern int qp_putse(PA char const * s) {
  E_print_escaped_str(CA s);
  return ctx->fail?-1:0;
}

extern int qp_putb(PA char const * s, size_t len) {
  E_write(CA s, len);
  return ctx->fail?-1:0;
}

extern int qp_putbe(PA char const * s, size_t len) {
  E_print_escaped_bytes(CA s, len);
  return ctx->fail?-1:0;
}

extern void qp_check(qprint_ctx * ctx) {
  if(ctx->fail) {
    qlog("qp_check: write(fd=%d) error %{ERR}",
        ctx->fd, (ctx->fail!=-1)?ctx->fail:0);
    exit(-1);
  }
}

